# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2018-09-12 08:07
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ldapauthbackend', '0002_ldapuserprofile_import_staff'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ldapuserprofile',
            name='import_staff',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist

from ldapauthbackend.backend import Connection, sanitize_username
from ldapauthbackend.models import LDAPUserProfile


def fetch_existed_user_dn():
	user_dn = {}
	dn_collection = set()
	for u in get_user_model().objects.all():
		try:
			if u.ldapuserprofile.dn:
				user_dn[u.username] = u.ldapuserprofile.dn
				dn_collection.add(u.ldapuserprofile.dn)
				continue
		except ObjectDoesNotExist:
			pass
		user_dn[u.username] = None
	return (user_dn, dn_collection)


def fetch_nonexist_candidate_members():
	conn = Connection.build_via_configuration()
	candidate_uids = conn.fetch_candidate_member_uids()
	existed_user, existed_dn = fetch_existed_user_dn()
	for n in existed_user.iterkeys():
		candidate_uids.discard(n)
	result = {}
	for u in candidate_uids:
		dn = conn.get_account_dn(u)
		if (not dn) or (dn in existed_dn):
			continue
		result[u] = dn
	return result


def import_candidate_member(user_name, user_dn, assoc_staff_user):
	user_name = sanitize_username(user_name)
	try:
		u = get_user_model().objects.get(username=user_name)
		raise ValueError("duplicated user: %r (given user-name=%r)" % (u, user_name))
	except ObjectDoesNotExist:
		pass
	conn = Connection.build_via_configuration()
	profile = conn.fetch_user_profile(user_dn)
	u = get_user_model().objects.create_user(user_name)
	ldap_profile = LDAPUserProfile(user=u, dn=user_dn, account_uid=profile.account_uid, account_name=profile.account_name, import_staff=assoc_staff_user)
	ldap_profile.save()
	return u


def sync_user_ldap_profile(user):
	conn = Connection.build_via_configuration()
	profile = conn.fetch_user_profile(user.ldapuserprofile.dn)
	user.ldapuserprofile.account_uid = profile.account_uid
	user.ldapuserprofile.account_name = profile.account_name
	user.save()
	return user



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib import admin
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.contrib.auth.models import User

from ldapauthbackend.models import LDAPUserProfile


class LDAPUserProfileInline(admin.StackedInline):
	model = LDAPUserProfile
	can_delete = False
	fk_name = 'user'
	verbose_name_plural = 'ldapuserprofile'


class UserAdmin(BaseUserAdmin):
	inlines = (LDAPUserProfileInline, )


# Re-register management classes

admin.site.unregister(User)
admin.site.register(User, UserAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from json import loads as json_loads

from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth import get_user_model
from django.contrib.auth.mixins import UserPassesTestMixin
from django.http import JsonResponse
from django.views.generic.base import View
from django.shortcuts import render

from ldapauthbackend.operations import fetch_nonexist_candidate_members
from ldapauthbackend.operations import import_candidate_member
from ldapauthbackend.operations import sync_user_ldap_profile


@staff_member_required
def index(request):  # pylint: disable=unused-argument
	return render(request, "ldapauthbackend/index.html")


class AccessCheckMixin(UserPassesTestMixin):
	raise_exception = True
	permission_required = None

	def test_func(self):
		u = self.request.user
		if not (u.is_active and u.is_staff):
			return False
		if self.permission_required:
			return u.has_perms(self.permission_required)
		return True


class CandidateMembers(AccessCheckMixin, View):
	def get(self, request, *args, **kwds):  # pylint: disable=unused-argument
		candidate_member_dn = fetch_nonexist_candidate_members()
		d = [{"username": u, "dn": d} for u, d in candidate_member_dn.iteritems()]
		d.sort(key=lambda x: x.get("username", None))
		result = {"candidates": d, "access_via": request.user.username}
		return JsonResponse(result)


class ImportMember(AccessCheckMixin, View):
	permission_required = ('auth.add_user', )

	def post(self, request, *args, **kwds):  # pylint: disable=unused-argument
		cmap = json_loads(request.body)
		username = cmap["username"]
		memberdn = cmap["memberdn"]
		u = import_candidate_member(username, memberdn, request.user)
		result = {
				"username": u.username,
		}
		return JsonResponse(result)


class SyncProfile(AccessCheckMixin, View):
	permission_required = ('auth.add_user', )

	def post(self, request, *args, **kwds):  # pylint: disable=unused-argument
		cmap = json_loads(request.body)
		username = cmap["username"]
		u = get_user_model().objects.get(username=username)
		u = sync_user_ldap_profile(u)
		result = {
				"username": u.username,
				"account_uid": u.ldapuserprofile.account_uid,
				"account_name": u.ldapuserprofile.account_name,
		}
		return JsonResponse(result)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from collections import namedtuple
from string import ascii_letters
from string import digits as ascii_digits
from json import load as json_load
import logging

from django.utils.encoding import force_text
from django.contrib.auth import get_user_model
from django.conf import settings as django_settings

import ldap

_log = logging.getLogger(__name__)

_mod = None
_cfg = None


class LDAPAuthBackendConfig(object):
	def __init__(self, urls, account_base_dn, candidate_group_dn, *args, **kwds):
		super(LDAPAuthBackendConfig, self).__init__(*args, **kwds)
		self.urls = urls
		self.account_base_dn = account_base_dn
		self.candidate_group_dn = candidate_group_dn

	@classmethod
	def load_configuration(cls):
		cfg_path = getattr(django_settings, "LDAP_AUTH_CONFIG_FILE")
		with open(cfg_path, "r") as fp:
			cmap = json_load(fp)
		urls = tuple(cmap["urls"])
		account_base_dn = tuple(cmap["account_base_dn"])
		candidate_group_dn = tuple(cmap["candidate_group_dn"])
		return cls(urls, account_base_dn, candidate_group_dn)


def get_config():
	# type: () -> LDAPAuthBackendConfig
	global _cfg
	if _cfg is None:
		_cfg = LDAPAuthBackendConfig.load_configuration()
	return _cfg


def get_module():
	# type: () -> ldap
	global _mod
	if _mod is None:
		mod_opt = getattr(django_settings, "LDAP_AUTH_MODULE_OPTIONS", None)
		if mod_opt:
			for opt_k, opt_v in mod_opt.iteritems():
				ldap.set_option(opt_k, opt_v)
		_mod = ldap
	return _mod


USERNAME_ACCEPT_CHAR = ascii_letters + ascii_digits + "_-."

UserProfile = namedtuple("UserProfile", ("dn", "account_uid", "account_name"))


def sanitize_username(n):
	return "".join([ch if (ch in USERNAME_ACCEPT_CHAR) else "_" for ch in n])


class ConnectError(Exception):
	def __init__(self, urls, *args, **kwds):
		super(ConnectError, self).__init__(*args, **kwds)
		self.urls = urls

	def __repr__(self):
		return "ConnectError(%r)" % (self.urls, )


_EMPTY_ITERABLE = ()


class Connection(object):
	def __init__(self, cfg, *args, **kwds):
		super(Connection, self).__init__(*args, **kwds)
		self.urls = cfg.urls
		self.account_base_dn = cfg.account_base_dn
		self.candidate_group_dn = cfg.candidate_group_dn
		self._link = None

	@classmethod
	def build_via_configuration(cls):
		# type: () -> Connection
		cfg = get_config()
		return cls(cfg)

	def _get_link(self):
		if self._link:
			yield self._link
		m_ldap = get_module()
		for u in self.urls:
			self._link = None
			l = m_ldap.initialize(u, bytes_mode=False)
			self._link = l
			yield l

	def _invoke(self, n, *args, **kwds):
		for lnk in self._get_link():
			c = getattr(lnk, n)
			try:
				r = c(*args, **kwds)
				return r
			except Exception:
				pass
		raise ConnectError(self.urls)

	def simple_bind(self, who, cred):
		return self._invoke("simple_bind_s", who, cred)

	def search(self, base, scope, filterstr='(objectClass=*)', attrlist=None, attrsonly=0, timeout=10):
		r = self._invoke("search_st", base, scope, filterstr, attrlist, attrsonly, timeout)
		if r:
			return r
		return _EMPTY_ITERABLE

	def user_authenticate(self, user_dn, password):
		try:
			self.simple_bind(user_dn, password)
			return True
		except Exception:
			_log.exception("failed on authorize user %r", user_dn)
		return False

	def get_account_dn(self, username):
		username = sanitize_username(username)
		filterstr = '(&(objectClass=account)(uid=' + username + '))'
		for base_dn in self.account_base_dn:
			r = self.search(base_dn, ldap.SCOPE_SUBTREE, filterstr, ('uid', ), 1)
			for dn, _entry in r:
				return dn

	def fetch_candidate_member_uids(self):
		result = set()
		for base_dn in self.candidate_group_dn:
			r = self.search(base_dn, ldap.SCOPE_SUBTREE, '(objectClass=posixGroup)', ('memberUid', ))
			for _dn, entry in r:
				l = entry.get('memberUid', None)
				if not l:
					continue
				result.update(l)
		return result

	def fetch_user_profile(self, user_dn):
		result = None
		r = self.search(user_dn, ldap.SCOPE_BASE, attrlist=('uid', 'uidNumber'))
		for dn, entry in r:
			user_name = entry['uid'][0]
			user_id = int(entry['uidNumber'][0])
			result = UserProfile(dn, user_id, user_name)
		if not result:
			raise KeyError("User DN not found: %r" % (user_dn, ))
		return result


UserModel = get_user_model()


class LDAPAuthBackend(object):
	def _can_authenticate(self, u):
		is_active = getattr(u, 'is_active', True)
		return is_active

	def _get_user_dn(self, username):
		if not username:
			raise ValueError('username must not empty')
		u = UserModel.objects.get(username=username)
		if not self._can_authenticate(u):
			raise ValueError('given user cannot authenticate')
		return (u, u.ldapuserprofile.dn)

	def authenticate(self, request, username=None, password=None):  # pylint: disable=unused-argument
		try:
			user_obj, user_dn = self._get_user_dn(username)
		except Exception:
			_log.exception("failed on get DN for user %r", username)
			return None
		if (not user_dn) or (not password):
			_log.error("empty user DN or not given password")
			return None
		conn = Connection.build_via_configuration()
		if conn.user_authenticate(force_text(user_dn), force_text(password)):
			return user_obj
		return None

	def get_user(self, user_id):
		try:
			return UserModel.objects.get(pk=user_id)
		except UserModel.DoesNotExist:
			pass
		return None



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.apps import AppConfig


class LDAPAuthBackendConfig(AppConfig):
	name = 'ldapauthbackend'
	verbose_name = 'LDAP Authentication Backend'



from django.conf.urls import url

from . import views

urlpatterns = [
		url(r'^$', views.index, name='index'),
		url(r'^candidate-members$', views.CandidateMembers.as_view(), name='candidate-members'),
		url(r'^import-member$', views.ImportMember.as_view(), name='import-member'),
		url(r'^sync-profile$', views.SyncProfile.as_view(), name='sync-profile'),
]



# -*- coding: utf-8 -*-
""" Setup script of LDAP authentication backend for Django """

from setuptools import setup

setup(
		name="django-ldap-auth-backend",
		version="0.1.6",  # REV-CONSTANT:rev 5d022db7d38f580a850cd995e26a6c2f
		description="LDAP authentication backend for Django",
		packages=[
				"ldapauthbackend",
				"ldapauthbackend.migrations",
		],
		include_package_data=True,
		install_requires=[
				"python-ldap >= 3.1.0, < 4.0.0",
				"django-angular-host-page-template-backend >= 0.1.0, < 2.0.0",
		])

# vim: ts=4 sw=4 ai nowrap



# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2018-09-19 06:33
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ldapauthbackend', '0003_auto_20180912_0807'),
    ]

    operations = [
        migrations.AddField(
            model_name='ldapuserprofile',
            name='account_name',
            field=models.CharField(default='', max_length=127),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='ldapuserprofile',
            name='account_uid',
            field=models.BigIntegerField(default=-1),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.db import models


class LDAPUserProfile(models.Model):
	user = models.OneToOneField(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
	dn = models.CharField(max_length=255)
	account_uid = models.BigIntegerField(default=-1)
	account_name = models.CharField(max_length=127)
	import_staff = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, null=True, related_name="+", editable=False)

	def __unicode__(self):
		return "LDAPUserProfile(user=%r, dn=%s, account_uid=%d, account_name=%r, import_staff=%r)" % (
				self.user,
				self.dn,
				self.account_uid,
				self.account_name,
				self.import_staff,
		)
